/******************************************************  DMA Functions  *******************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to control and monitor the DMA engines.

	Rev History: 11-28-2018:	Inital release.

	Copyright 2018, Red Rapids, Inc.

******************************************************  DMA Functions  *******************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "RRadapter.h"
#include "sigfpga.h"
#include "dma_functions.h"

// Assign a DMA buffer to a device channel.
// Return of zero indicates success, non-zero indicates an error condition.
int DMAAssign(s_Handle *p_HW, unsigned BufferIndex)
{
	unsigned pagecnt;
	unsigned long long descriptor;
	unsigned addindex;
	unsigned long long *physical_address;
	
	// Allocate memory for the physical addresses.
	if ((physical_address = malloc(sizeof *physical_address * p_HW->DMA[BufferIndex].PageCount)) == NULL)
		return errno;

	// Allocate a DMA buffer for the requested index.
	if (DMABufAllocate(&p_HW->DMA[BufferIndex], p_HW->DevNum, physical_address))
	{
		free(physical_address);
		return DMA_ALLOCATE;
	}

	// Set the DMA descriptors register index to load page offsets for the selected channel.
	Write64(p_HW, BAR0, DMA_DESCRIPTORS - 0x8, 0x2 << BufferIndex);
	
	// Initialize the address index to the start of the page memory.
	addindex = DMA_DESCRIPTORS;
			
	// Write the DMA physical page addresses to the adapter.
	for(pagecnt=0; pagecnt < p_HW->DMA[BufferIndex].PageCount; pagecnt++)
	{
		// Perform the write operation to the page memory.
		Write64(p_HW, BAR0, addindex, physical_address[pagecnt]);
		// Increment the page memory address.
		addindex += 8;
	}
	
	// Set the DMA descriptors register index to load descriptors for the selected channel.
	Write64(p_HW, BAR0, DMA_DESCRIPTORS - 0x8, 0x200 << BufferIndex);
			
	// Initialize the address index to the start of the descriptor memory.
	addindex = DMA_DESCRIPTORS;
	
	// Form the descriptors and write them to the adapter.
	for(pagecnt=0; pagecnt < p_HW->DMA[BufferIndex].PageCount; pagecnt++)
	{
		// Set the Burst Size and Burst Count fields of the descriptor.
		descriptor = (unsigned long long)(p_HW->DMA[BufferIndex].BurstSizeB << 16) | (p_HW->DMA[BufferIndex].BurstCount);
	
		//Set the Last Page descriptor bit if this is the final page of this channel.
		if (pagecnt == p_HW->DMA[BufferIndex].PageCount - 1)
			descriptor |= 0x0000000040000000ull;
	
		//Set the Interrupt descriptor bit if this page matches the requested pages per interrupt setting.
		if (((pagecnt+1) % p_HW->DMA[BufferIndex].PagesPerMark) == 0)
			descriptor |= 0x0000000080000000ull;
		// Perform the write operation to the descriptor memory.
		Write64(p_HW, BAR0, addindex, descriptor);
		// Increment the descriptor memory address.
		addindex += 8;
	}

	free(physical_address);
	return OK;
}

// Verify that the DMA configuration settings have legal values.
// Return of zero indicates no errors, non-zero indicates an error condition.
int DMAConfigCheck(s_DMABuf DMA, unsigned *ConfigError)
{
	unsigned pagecnt;
	unsigned index = 0;
	
	for(pagecnt=0; pagecnt < DMA.PageCount; pagecnt++)
	{
		// Determine whether the burst size is within range and an even multiple of the byte alignment.
		if ((DMA.BurstSizeB > MAX_BURSTSIZEB) || fmod(DMA.BurstSizeB, BYTE_ALIGNMENT)) ConfigError[0]++;

		// Determine whether the burst count is within range.
		if (DMA.BurstCount > MAX_BURSTCOUNT) ConfigError[1]++;
	}

	// Determine whether the page count per channel is within range.
	if (DMA.PageCount > MAX_PAGECOUNT) ConfigError[2]++;

	// Determine whether the PagesPerMark divides evenly into the PageCount.
	if (fmod(DMA.PageCount, DMA.PagesPerMark)) ConfigError[3]++;
		
	for (index = 0; index < 4; index++)
		if (ConfigError[index]) return DMA_BUFFER;

	return OK;
}

// Initializes all varibles in the s_DMABuf structure to default values.
void DMAInit(s_DMABuf *p_DMA)
{
	// Set all of the size and count variables to zero.
	p_DMA->PageNumber = 0;
	p_DMA->PageCount = 0;
	p_DMA->PagesPerMark = 0;
	p_DMA->BurstSizeB = 0;
	p_DMA->BurstCount = 0;
	// Set all pointers to NULL.
	p_DMA->Address = NULL;
}

// Preload a transmitter buffer with sine wave data samples.
// Return of zero indicates that the buffer contains an integer number of sine wave periods, non-zero indicates a fractional number of periods.
int DMALoad(char *p_File, s_DMABuf *DMA, unsigned DataItemSize)
{
	FILE *fp;
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	unsigned page_count = 0;
	unsigned record_count = 0;
	unsigned data_items_count = 0;
	int sample = 0;

	// Sequence through all of the number of pages requested.
	for(page_count = 0; page_count < DMA->PageCount; page_count++)
	{
		// Open the requested file and read enough samples to fill a single DMA page.
		if((fp = fopen(p_File, "r")) == NULL)
			return errno;
		else
		{
			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < (DMA->BurstSizeB * DMA->BurstCount) / 8; record_count++)
			{
				// Clear the current 64-bit record.
				DMA->Address[page_count][record_count] = 0x0ull;
				
				for(data_items_count = 0; data_items_count < 64 / DataItemSize; data_items_count++)
				{
					// Read the next sample from the file.
					if (fscanf(fp, "%d", &sample) != 1)
						return EIO;
					
					// Write the sample as the current data item to the payload record.
					DMA->Address[page_count][record_count] |= ((max >> (64-DataItemSize)) & (unsigned long long)sample) << data_items_count*DataItemSize;
				}
			}
		// Close the file.
		fclose(fp);
		}
	}
	return OK;
}

// Release allocated DMA buffers.
// Return of zero indicates success, non-zero indicates an error condition.
int DMARelease(s_DMABuf *DMA, unsigned DevNum)
{
	// Free the requested DMA buffer.
	return(DMABufFree(DMA, DevNum));
}

// Write the contents of the DMA buffer to the specified file from the first page to the number of pages requested.
// Return of zero indicates success, non-zero indicates an error condition.
int DMASave(char *p_File, s_DMABuf *DMA, unsigned NumPages, unsigned DataItemSize)
{
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	unsigned page_count = 0;
	unsigned data_items_count = 0;
	unsigned record_count = 0;
	signed sample = 0;

	// Open the requested file and process samples from the DMA buffer.
	if((fp = fopen(p_File, "w")) == NULL)
		return errno;
	else
	{
		// Sequence through all of the DMA pages assigned to the channel.
		for(page_count = 0; page_count < NumPages; page_count++)
		{
			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < (DMA->BurstSizeB * DMA->BurstCount) / 8; record_count++)
				// Sequence through every sample in the 64-bit record.
				for (data_items_count = 0; data_items_count < 64 / DataItemSize; data_items_count++)
					{
						// Extract the next sample value from the 64-bit record.
						sample = (signed)(DMA->Address[page_count][record_count] >> (data_items_count * DataItemSize)) & (max >> (64 - DataItemSize));
						// Sign extend negative numbers.
						if (sample & (one << (DataItemSize - 1))) sample |= (max << DataItemSize);
						// Print the sample value to the file.
						fprintf(fp, "%d\n", sample);
					}
		}
		// Close the file.
		fclose(fp);
	}
	return OK;
}

// Report the current DMA on Demand operating status.
void GetDMAStatus(s_Handle *p_HW, unsigned ChannelNumber, s_DMAStatus *p_DMAStatus)
{
	unsigned long long value = 0;
	unsigned cnt = 0;
	
	// Read the requested DMA status address.
	switch (ChannelNumber)
	{
	case 1: {Read64(p_HW, BAR0, DMA_1TO2_STATUS, &value); break;}
	case 2: {Read64(p_HW, BAR0, DMA_1TO2_STATUS, &value); value >>= 32; break;}
	case 3: {Read64(p_HW, BAR0, DMA_3TO4_STATUS, &value); break;}
	case 4: {Read64(p_HW, BAR0, DMA_3TO4_STATUS, &value); value >>= 32; break;}
	case 5: {Read64(p_HW, BAR0, DMA_5TO6_STATUS, &value); break;}
	case 6: {Read64(p_HW, BAR0, DMA_5TO6_STATUS, &value); value >>= 32; break;}
	case 7: {Read64(p_HW, BAR0, DMA_7TO8_STATUS, &value); break;}
	case 8: {Read64(p_HW, BAR0, DMA_7TO8_STATUS, &value); value >>= 32; break;}
	default: break;
	}

	// Strip the current page.
	p_DMAStatus->CurrentPage = value & 0x00000000000003FFull;

	// Strip the current burst.
	p_DMAStatus->CurrentBurst = (value >> 16) & 0x000000000000FFFFull;

	// Strip the individual DMA engine status bits (0 = Enable, 1 = Go, 2 = Busy).
	for (cnt = 0; cnt < 2; cnt++)
		p_DMAStatus->Engine[cnt] = (value >> (cnt + 10)) & 0x0000000000000001ull;
}
